/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 1995,2012          */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
/*
 *   File name: mcpgath.c
 *
 * History:
 *      03/27/98 AJI M01 changes to #define BLOCK_SIZE (D38675)
 *      05/19/98 AJI M02 changes to dynamic blocksize (D39519)
 *      06/17/98 PFC M03 fix labels on messages added by defects 38675 and
 *                       39519 (D41184)
 *      06/18/98 SVC M04 stdlib.h needed for getenv() (D40375)
 *      07/06/98 AJI M05 Routine needs multiple enhancements from counting # of flags 
 *                       and multiple messages issued when file already exists (D34888)
 *      07/16/98 AJI M06 mcpgath hangs when started with one task (D41944)
 *      02/03/99 PFC M07 check for DFS f/s when checking space (D46200)
 *      03/11/99 SVC M08 OS/390: move to mohonk, undo M07 (no DFS for OS/390) (D47705)
 *      04/20/05 ECL M09 Liux port (F118848/D120380)
 *      04/02/06 PFC M10 Initialize glob data areas for Linus (D134398)
 *
 *   Origins:
 *                   Copyright IBM Corporation 1995, 2002
 *
 *                         All Rights Reserved
 *
 *   Permission to use, copy, modify, and distribute this software and its
 *   documentation for any purpose and without fee is hereby granted,
 *   provided that the above copyright notice appear in all copies and that
 *   both that copyright notice and this permission notice appear in
 *   supporting documentation, and that the name of IBM not be
 *   used in advertising or publicity pertaining to distribution of the
 *   software without specific, written prior permission.
 *
 *   Information: mcpgath does a parallel file copy gather, taking a number    
 *   of files from tasks 0 thru N and copies them in sequence to task 0.
 *   mcpgath expects that each file is to exist on each task, and will gather
 *   them back to task 0.  
 * 
 *   mcpgath is expected to be used to copy multiple versions of a file from
 *   multiple tasks back to a single task.

 *   The syntax is:
 *
 *      mcpgath [-ai] source  .... destination  [POE options]   where
 *
 *         -a is an optional flag when specified denotes to append the task number
 *            to the end of the file name when it is copied to task 0.  This is for
 *            task identification purposes to know where the data came from.  The -a
 *            and -i flags can be combined to check for existing files appended with
 *            the task number.
 *
 *         -i is an optional flag when specified denotes to check for duplicate
 *            or existing files of the same name, and to not replace any existing
 *            file found.  If this flag is not specified, the default action is
 *            to replace any existing files with the source file. The -a
 *            and -i flags can be combined to check for existing files appended with
 *            the task number.
 *
 *         source  is one of the following:
 *            - one or more existing file names, in which case these files will be
 *              copied with the same names to the destination directory on task 0. 
 *              Each file name specified must exist on all tasks involved in the copy.
 *            - a directory name, in which case all files in that directory on each task
 *              are copied with the same names to the destination directory on task 0.
 *            - an expansion of file names, using wildcards, in which case those files
 *              are copied with the same names to the destination directory.  All
 *              wildcarded input strings must be enclosed in double quotes.
 *
 *         destination  is an existing destination directory name where the data will be
 *            copied to.
 *
 *         POE options  are any valid POE command line argument.  You can also
 *            specify any valid POE environment variable.
 *
 *   The number of nodes to move the file to (-procs), and the message
 *   passing protocol used to move the file (-euilib) are the POE
 *   options which may be of most interest to the user. See the IBM
 *   Parallel Environment for AIX Operation and Use guide for 
 *   information on POE options.
 *
 *   A default of 100K data block size is used for copying the data.  This can be
 *   changed by setting the enviornment variable MP_BLKSIZE between 1 and 8000000.
 *
 *   return codes:  -1 : invalid number of arguments specified
 *                  -2 : invalid option flag specified      
 *                  -3 : unable to resolve input file name(s)       
 *                  -4 : Could not open input file for read                  
 *                  -5 : no room on destination node's filesystem                 
 *                  -6 : error opening file output file    
 *                  -7 : error creating output file       
 *                  -8 : error writing to output file         
 *                  -9 : MPI_Send of data failed              
 *                  -10 : Final MPI_Send failed              
 *                  -11 : MPI_Recv failed              
 *                  -12 : invalid block size
 *                  -13 : error allocating storage 
 *                  -14 : total # of tasks must be greater than 1
 *
 *
 */

#pragma comment(copyright,"@(#) 5765-F83 (C) Copyright IBM Corp. 1995,2002\0")
#pragma comment(user,"@(#) 1.26 src/ppe/poe/samples/mpi/mcpgath.c, ppe.poe.samples, ppe_rbarlx2, rbarlx21216b 12/04/23 13:26:33")

#include <stdio.h>
#include <stdlib.h>      /* M04 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#ifndef __linux   /* M09: sys/vnode.h not available on Linux */
 #include <sys/vnode.h>
#endif
#include <glob.h>
#include <mpi.h>

#define MAX_BLOCK_SIZE 8000000  /* M01 maximum block size user can request */

main(argc,argv)
int argc;
char *argv[];
{
  glob_t globbuf;              /* structure for obtaining file names */
  char in_file[FILENAME_MAX];  /* name of file to be moved */
  char *dest_dir;              /* name of output destination directory */
  FILE *fp;                    /* file pointer for input file */
  FILE *fp_out;                /* file pointer for file being written */
  struct stat statb;           /* structure needed for stat() function */
  struct statvfs stata;        /* structure needed for statvfs() function */
#ifndef __linux                  /* M09: VDIR & st_type not available on Linux */
  #define STATB_STTYPE_IS_VDIR statb.st_type == VDIR
#else
  #define STATB_STTYPE_IS_VDIR S_ISDIR(statb.st_mode)
#endif
  int i,j,k,n;                 /* general purpose integers */
  int opened_file;             /* indicates if out file is open for writing */
  int c;
  int fd;
  int source=0,type=0,msg_rc;
  MPI_Status status;

  struct file_def {            /* file data sent to each node */
    int  name_size;            /* size of file name */
    long  file_size;            /* size of the file */
    int  content_size;         /* size of file data in this send */
    char name[FILENAME_MAX];   /* file name */
    mode_t file_mode;          /* file mode (permisssions etc...) */
  } file_definition;

  struct file_tab {            /* table of open output files  */
    char name[FILENAME_MAX];   /* file name */
    FILE *ptr;                 /* file pointer */
 } file_tbl[FOPEN_MAX], save_filename[FOPEN_MAX];  /* M05 */
 
  char task0_filename[FILENAME_MAX]; /* Task 0 output file name */
  char *in_name; /* input file name */ 

  char *env_block_ptr;         /* M01 pointer to env var to get MP_BLKSIZE value */
  char *send_buffer;           /* M02 pointer to send_buffer area */ 
  char *rec_buffer;            /* M02 pointer to rec buffer area */
  struct file_def *fdp;        /* M02 */
  char *contentps;             /* M02 pointer to send file data  */
  char *contentpr;             /* M02 pointer to receive file data */
 
  long int blocksize = 100000; /* M01 default blocksize is used if MP_BLKSIZE is not set */

  int number_of_tasks;         /* # nodes (including sender) executing */
  int my_task_number;          /* task number of this node */
  int numfiles = 0;
  int prev_numfiles = 0;
  int exist = 0;      
  int append = 0;      
  int lasttype = 41;
  int tasks;
  char cbuf[5];
  int first_time_send_buffer = 1;    /* M02 */
  int first_time_rec_buffer = 1;     /* M02 */
  int print_msg = 0;                 /* M05 */
                
/***********************************************/
/* Set up message passing environment, using   */
/* the number of tasks (procs) to set up the  */
/* task group.                                */
/***********************************************/

MPI_Init(&argc,&argv); /* initialize MPI environment */
MPI_Comm_rank(MPI_COMM_WORLD, &my_task_number); /* get task number */
MPI_Comm_size(MPI_COMM_WORLD,&number_of_tasks); /* get number of tasks */

if (number_of_tasks <= 1) {       /* M06 number of -procs must be more than 1 */
   fprintf(stderr,"mcpgath: total number of tasks must be greater than 1 \n");
                                  /* M06 if not, issue error message */
   MPI_Abort(MPI_COMM_WORLD,-14); /* M06 stop all other tasks */
}
/***********************************************/
/* Parse arguments and check for -i flag       */
/***********************************************/

while (--argc > 0 && (*++argv)[0] == '-')
  while (c = *++argv[0])  /* an option was found ... */
     switch (c) {
       case 'a':
           append = 1;  /* append task number to file name */
           break;
       case 'i':
           exist = 1;   /* check for duplicate existing files */
           break;
       default:
           fprintf(stderr,"mcpgath:  Illegal option %c \n", c);
           fprintf(stderr,"usage: mcpgath [-ai] source destination\n"); 
           MPI_Abort(MPI_COMM_WORLD,-2); /* stop all other tasks */
     }

if(argc < 2 ) {    /* M05 should be 2-n arguments to this prog */
      fprintf(stderr,"usage: mcpgath [-ai] source ... destination\n"); 
                                  /* if not, issue error message */
      MPI_Abort(MPI_COMM_WORLD,-1); /* stop all other tasks */
}

/* Handle destination directory, since that must be the last parameter */

dest_dir = (char *) malloc(strlen(argv[argc-1])+1);   /*stg for output dir name*/
strcpy(dest_dir,argv[argc-1]);                      /* get output dir name */

/* Read in first source input string...there may be several source files */
/* The source input string may also be a directory, instead of a file.   */

strcpy(in_file,argv[0]);                      

/* Use a stat() call to see if the input string may be a directory.  If it */
/* is, tack on a slash and wildcard character to allow glob() to resolve   */
/* file names in that directory.                                           */

stat(in_file,&statb);  
if ( STATB_STTYPE_IS_VDIR ) strcat(in_file,"/*");

/* Make an initial glob() call to return the file name(s) of the first  */
/* source input string.                                                 */

/* Source input strings can either be an explicitly specified name of     
   one or more files, a directory, or a "fuzzy" representation of files 
   using meta characters (such as *.o) as long as the string is enclosed 
   in double quotes.                                                     */

#ifdef __linux
globbuf.gl_pathc = 0; /* M10 initialize glob data areas before call for Linux */
globbuf.gl_pathv = NULL;
globbuf.gl_offs = 0;
#endif
glob(in_file,GLOB_DOOFFS,NULL,&globbuf);
if (globbuf.gl_pathc <= 0) {   /* glob failed  */
   fprintf(stderr,"mcpgath:  Unable to resolve list of files for %s.\n",in_file);
   MPI_Abort(MPI_COMM_WORLD,-3); /* stop all other tasks */
}

/* If there are a number of source input files specified, use a loop to */
/* do more glob() calls to resolve those file names.                    */

for (i=0; i < argc-2; i++) {    
   strcpy(in_file,argv[i+1]);                      
   stat(in_file,&statb);  /* Check to see if source is a directory */
   if ( STATB_STTYPE_IS_VDIR ) strcat(in_file,"/*");
   prev_numfiles = globbuf.gl_pathc;
   glob(in_file,GLOB_APPEND|GLOB_DOOFFS,NULL,&globbuf);
   if (globbuf.gl_pathc == prev_numfiles ) {   /* glob failed  */
      fprintf(stderr,"mcpgath:  Unable to resolve list of files for %s.\n",in_file);
      MPI_Abort(MPI_COMM_WORLD,-3); /* stop all other tasks */
   }
}
numfiles = globbuf.gl_pathc;
i = 0;
/***********************************************************************/
/*                                                                     */
/*  The following is the major loop for gathering the files.  It is    */
/* governed by the total number of files there is to be gathered.  For */
/* each file on task 0, the file will be copied directly to the desti- */
/* nation directory.  For files on tasks 1 thru N, the file is sent    */
/* via an MPI blocking send and receive to task 0.  Files that do not  */
/* exist or cannot be accessed are skipped and the remaining files are */
/* copied.                                                             */
/*                                                                     */
/***********************************************************************/
while(i < numfiles) { /***  BEGINNING OF LOOP FOR SENDING FILES  ***/

    file_definition.name_size = strlen(globbuf.gl_pathv[i]); /* get len file */
    strcpy(file_definition.name,globbuf.gl_pathv[i]); /* get name of file */

    /***********************************************/
    /* if no information is available about the    */
    /* file or the file cannot be opened...        */
    /***********************************************/

    if(stat(globbuf.gl_pathv[i],&statb) == -1 || ! (fp=fopen(globbuf.gl_pathv[i],"r"))) {
      fprintf(stderr,"mcpgath: Could not open file %s for reading\n",globbuf.gl_pathv[i]);
      MPI_Abort(MPI_COMM_WORLD,-4); /* stop all other tasks */
    }

    /* Check to see if file is a directory.  If so, it must be skipped. */

    if ( STATB_STTYPE_IS_VDIR ) {
      i++;   /* process next file  */
      fclose(fp);  /* close this file  */
      continue; /* continue with next file */
    }

    if (statb.st_size == 0) {     /* File is empty....issue msg */
      fprintf(stderr,"mcpgath: File %s is empty\n",globbuf.gl_pathv[i]);
      i++;
      fclose(fp);
      continue; /* continue with next file */
    }

    file_definition.file_mode = statb.st_mode;  /* get file mode */
    file_definition.file_size = statb.st_size;  /* get file len from struct */

    /* The destination output file name will be contstructed by taking  */
    /* the source file name, stripping off its path name, and appending */
    /* it to the destination directory name.                            */

    strcpy(task0_filename,dest_dir);  
    strcat(task0_filename,"/"); 
    in_name = (char *)strrchr(file_definition.name,'/');  /* Strip off path name */
    in_name ? strcat(task0_filename,++in_name) : 
              strcat(task0_filename,file_definition.name);  

    /* When the -a flag is specified, the task number needs to be       */
    /* appended to the output file name.                                */

    if (append == 1) { 
            strcat(task0_filename,".");  
            sprintf(cbuf,"%d\0",my_task_number);
            strcat(task0_filename,cbuf);  
    }
                          
    /***********************************************/
    /* Check to see if env varible is set thru cmd */
    /* line. If the environment variable was set   */
    /* get the value of it. Check to see if block  */
    /* size is with in limits, if so allocate the  */
    /* storage based on the MP_BLKSIZE.            */
    /***********************************************/

      if (first_time_send_buffer == 1)  {  /*   M02 allocate the data area once for every send */
  
         if ((env_block_ptr = getenv("MP_BLKSIZE")) != NULL) { /*  M01 */
            blocksize = atoi(env_block_ptr);  /*  M01 */
          } /*  M01 */
         if ((blocksize <= 0) || (blocksize > MAX_BLOCK_SIZE)) {  /*   M01 */
            fprintf(stderr, "mcpgath: blocksize has to be between 1 and %ld\n", MAX_BLOCK_SIZE);  /* M01 */
                     /* M03 added function label to message */
            MPI_Abort(MPI_COMM_WORLD,-12); /*  M01 stop all other tasks */
          }   /*  M01 */
 
        /* M01 Allocate data area based on block size */

        if ((send_buffer = (char *) malloc(sizeof(file_definition) + blocksize)) == NULL) { /* M02 */
           fprintf(stderr,"mcpgath:  error allocating storage for send buffer: \n"); /* M02 */
                     /* M03 added function label to message */
           MPI_Abort(MPI_COMM_WORLD,-13); /* M02 stop all other tasks */
         }   /* M02 */
        first_time_send_buffer += 1;  /* M02 */
      }  /* M02 */
  
     contentps = send_buffer + sizeof(file_definition);  /* M02 */

    opened_file = 0;          
    if(my_task_number == 0)    /***********************************************/
                               /* If this is Task 0, there is no need to send */
                               /* the file.  We can write it directly to its  */
                               /* destination directory and not bother with   */
                               /* message passing.                            */
                               /***********************************************/

        if (opened_file == 0) {  /* Output file not open yet.... */
 
             /************************************************/
             /* Verify that there is space for the file.     */
             /* Divide the file size by the block size of    */
             /* the file system. If that value is > the      */
             /* number of free blocks in the file system,    */
             /* stop processing this task and return.        */
             /************************************************/
             
            strcpy(file_definition.name,task0_filename); 
            if(!(statvfs(file_definition.name, &stata))) {
                   /* file exists ... if -i flag specified, we do not want to replace it.     */
                   /* Do not replace existing file, issue msg and continue */

                    if (exist == 1) {    
                       fprintf(stderr,
                       "mcpgath: File %s already exists on task 0, and will not be copied.\n",
                       file_definition.name);
                       i++;
                       continue;          /* continue with next file   */
                    }

      /* M07 Check file system space only if file sys type is not DFS */

#ifndef __linux  /* M08/M09: No DFS support on  Linux */
                  if (strncmp(stata.f_basetype,"dfs",3) != 0)    /* M07 */
#endif       /* M08 */
                    if (stata.f_bavail<((unsigned long)file_definition.file_size/stata.f_frsize+1)) {
                           fprintf(stderr,"mcpgath:  no room in filesystem to copy %s for task %d.\n", 
                                   file_definition.name, my_task_number);
                           /* M03 fixed spacing in message */
                           MPI_Abort(MPI_COMM_WORLD,-5); /* stop all other tasks */
                     }

                    if( !(fp_out = fopen(file_definition.name,"w")) ) { /* open "to" file */
                         fprintf(stderr," mcpgath: Failed opening %s on task %d\n",
                               file_definition.name,my_task_number);
                         MPI_Abort(MPI_COMM_WORLD,-6); /* stop all other tasks */
                    }
                    opened_file = 1;         /* "to" file on originating opened */
            }
            else {                  /* statvfs failed, file must not exist,
                                       so create it. */
                  fd = creat(file_definition.name,file_definition.file_mode);
                  fp_out = fdopen(fd,"w");
                  if (fp_out == NULL) {   /* open failed.... */
                        fprintf(stderr,"mcpgath: Failed create of output file %s for task %d \n",
                                 file_definition.name,status.MPI_SOURCE);
                        MPI_Abort(MPI_COMM_WORLD,-7); /* stop all other tasks */
                  }
                  opened_file = 1;           /* "to" file on originating opened */
            }
         }        /* end if task = 0, opening file for write */

    
    do {     /* read contents of the file */

        if(statb.st_size > blocksize) {  /* M01 if # bytes left in file > block */
          k = blocksize;                 /* M01 write amount = 1 block */
        }
        else k = statb.st_size; /* if # bytes left < or = block */
                                /* write amount = amount left */
        statb.st_size -= k;              /* decrement amount left to write */

        if( !fread(contentps,sizeof(char),k,fp)) { /* M02 read piece */
          fprintf(stderr,"mcpgath Attention: fread of %s may have failed\n",
                  file_definition.name);
        }
        file_definition.content_size = k;  /* # bytes read in structure sent */
   
        if(my_task_number == 0) {  /* if this is task 0 ..... copy file directly */

          if((j=fwrite(contentps,sizeof(char),k,fp_out))!=k) {  /* M02 */
               fprintf(stderr," mcpgath: Failed writing of %s on task %d\n",
               file_definition.name,my_task_number); /* write failure */
               fprintf(stderr, "mcpgath: Wrote %d bytes while writing %d\n",j,k);
               MPI_Abort(MPI_COMM_WORLD,-8); /* stop all other tasks */
          }
        }
        else  {  /* reading a file for task 1 thru task N, send it to that task... */

           strcpy(file_definition.name,task0_filename); 
           memcpy(send_buffer, (char *) &file_definition, sizeof(file_definition));
                                                      /* M02 prepare for send_buffer */
           msg_rc = MPI_Send(send_buffer, sizeof(file_definition)+blocksize,MPI_CHAR,
                    source,type,MPI_COMM_WORLD);   /* M02 */
           if (msg_rc) {
               fprintf(stderr," mcpgath: MPI_Send of %s for task %d failed.\n",
                       file_definition.name,my_task_number); 
               MPI_Abort(MPI_COMM_WORLD,-9); /* stop all other tasks */
           }
         }
    } while (statb.st_size > 0);   /* doing read of the input file */
    i++;
    fclose(fp);                      /* close the file we've been reading */
    if (my_task_number == 0) {  /* Task 0, we copied files directly */
      fclose(fp_out);                /* close the file we've been writing */
      chmod(file_definition.name,file_definition.file_mode); /* set file mode */
      opened_file = 0;
    }
}   /* end of loop sending files */
if (my_task_number != 0) {  /* Task 1 thru N, send a final msg to task 0 */
  msg_rc = MPI_Send(&i,sizeof(i),MPI_CHAR,0,lasttype,MPI_COMM_WORLD);
  if (msg_rc) {
      fprintf(stderr," mcpgath: last MPI_Send to task %d failed.\n", 
              my_task_number); 
      MPI_Abort(MPI_COMM_WORLD,-10); /* stop all other tasks */
  }
}    /* end sending final msg for tasks 1 thru N */


/************** END OF SEND LOGIC ***********************************************/
    
if (my_task_number == 0) {  /* Task 0, we have files to receive */


/******************************************************/
/* This is the receive side logic, for receiving      */
/* the files, from tasks 1 thru N.  Task 0 is handled */
/* on the send side, by writing the file directly to  */
/* the destination.                                   */
/* Task 0 collects all of the files.  Here, we will   */
/* have to do a series of MPI_Recv calls to accept    */
/* incoming data.  The data can come from any and all */
/* tasks, and in the case of copying large amounts    */
/* of data, may also be coming to Task 0 as part      */
/* of several blocks.  For this reason, we need to    */
/* keep track of the file the data is to be written   */
/* to, using a table of file pointers and names.      */
/*                                                    */
/* We will continue to receive data until we have     */
/* received the final message from each task.         */
/******************************************************/
    
          /* Initialize the file pointers in the table to null,
             to show no files are open yet.   */
  for (i = 0; i < FOPEN_MAX-1; i++) file_tbl[i].ptr = NULL;
  tasks = 1;
  do {   /* Receive messages until we get a message of a particular type, that 
            signifies the last file has been sent. */
        if (first_time_rec_buffer == 1) {  /*   M02 recieve side buffer allocation */
           if ((rec_buffer = (char *) malloc(sizeof(file_definition) + blocksize)) == NULL) {
                            /* M02  allocate the storage based on file_definition & blocksize  */
              fprintf(stderr,"mcpgath:  error allocating storage for receive buffer  \n"); /* M02 */
                     /* M03 added function label to message */
              MPI_Abort(MPI_COMM_WORLD,-13); /* M02 stop all other tasks */
            }   /* M02 */
           first_time_rec_buffer +=1;  /* M02 */
         }
        fdp = (struct file_def *)rec_buffer;  /* M02 */
        msg_rc = MPI_Recv(rec_buffer,sizeof(file_definition)+blocksize,MPI_CHAR,
                 MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&status);   /* M02 */
 
        if (msg_rc) {    /* M02  */
           fprintf(stderr," mcpgath: MPI_Recv of %s from task %d failed.\n",
           fdp->name,status.MPI_SOURCE); /*  M02 write failure */
           MPI_Abort(MPI_COMM_WORLD,-11); /* M02 stop all other tasks */
         }
        contentpr = rec_buffer + sizeof(file_definition);  /* M02 start of content location */   
        if (status.MPI_TAG == lasttype) {       
               /* Last message received, break out of loop   */
               /* when all the tasks have sent this msg.     */
           tasks++;  
           if (tasks == number_of_tasks) break;  
           continue;
        }
        opened_file = 0;
        for (i = 0; i < FOPEN_MAX-1; i++)  {
           if (strcmp(file_tbl[i].name,fdp->name) == 0) {  /* M02 */
                fp_out = file_tbl[i].ptr; 
                opened_file = 1;   /* The data received is for a file that is */
                break;             /* already open.                           */
           }
        }

        if (opened_file == 0) {
    
           /************************************************/
           /* Verify that there is space for the file.     */
           /* Divide the file size by the block size of    */
           /* the file system. If that value is > the      */
           /* number of free blocks in the file system,    */
           /* stop processing this task and return.        */
           /************************************************/
    
            if(!(statvfs(fdp->name, &stata))) {  /* M02 */ 
                /* file exists ... if -i flag specified, we do not want to replace it.  */
                /* Do not replace existing file, issue msg and continue */

                if (exist == 1) {   
                   for (i = 0; i < FOPEN_MAX-1; i++)  { /* M05 */
                       if (strcmp(save_filename[i].name,fdp->name) == 0) {  /* M05 */
                          print_msg = 0;  /* M05  no need to issue message */
                          break; /* M05 no need to look futher  */
                       }  /* M05 */ 
                       else print_msg = 1; /* M05 set up flag to issue message */
                   }  /* M05 */
                   if (print_msg == 1) {  /* M05 issue the message */  
                       fprintf(stderr,
                             "mcpgath: File %s already exists on task %d, and will not be copied.\n",
                             fdp->name,my_task_number);  /* M02 */
                       print_msg = 0;  /* M05 */
                       for (i = 0; i < FOPEN_MAX-1; i++)  {
                           if (strcmp(save_filename[i].name,"") == 0) {  /* M05 find next empty place */ 
                              strcpy(save_filename[i].name,fdp->name);  /* M05 save the file name there */
                              break;  /* M05 no need to look further */
                           }  /* M05 */
                       } /* M05 */
                   }  /* M05 */
                   continue;          
                }


      /* M07 Check file system space only if file sys type is not DFS */

#ifndef __linux  /* M08/M09: No DFS support on Linux */
              if (strncmp(stata.f_basetype,"dfs",3) != 0)    /* M07 */
#endif       /* M08 */
                if (stata.f_bavail<((unsigned long)fdp->file_size/stata.f_frsize+1)) {  /* M02 */
                    printf("mcpgath:no room in filesystem to copy %s for task %d.\n", 
                            fdp->name, my_task_number);  /* M02 */
                    MPI_Abort(MPI_COMM_WORLD,-5); /* stop all other tasks */
                }
                if( !(fp_out = fopen(fdp->name,"w")) ) { /*  M02 open existing file */
                      fprintf(stderr,"mcpgath: Failed opening output file %s for task %d \n",
                               fdp->name,status.MPI_SOURCE);
                      MPI_Abort(MPI_COMM_WORLD,-6); /* stop all other tasks */
                }
            }  /* end statvfs successful, file exists. */

            else {           /* statvfs failed, file must not exist, so create it. */

               fd = creat(fdp->name,fdp->file_mode);  /* M02 */
               fp_out = fdopen(fd,"w");
               if (fp_out == NULL) {   /* open failed.... */
                    fprintf(stderr,"mcpgath: Failed create of output file %s for task %d \n",
                             fdp->name,status.MPI_SOURCE);  /* M02 */
                    MPI_Abort(MPI_COMM_WORLD,-7); /* stop all other tasks */
               }
               opened_file = 1;    

            }  /* end statvfs failed; file does not exist */
            for (i = 0; i < FOPEN_MAX-1; i++)  {
               if (file_tbl[i].ptr == NULL) {
                 file_tbl[i].ptr = fp_out;  /* Save ptr to output file in table */
                 strcpy(file_tbl[i].name,fdp->name);  /* M02 */
                 break;
               }
            }

      }    /* end if output file has not been opened.... */


  /* Check filesystem size to ensure there is enough room to hold data */

    statvfs(fdp->name, &stata);  /* M02 */
    if (stata.f_bavail<((unsigned long)fdp->file_size/stata.f_frsize+1)) {  /* M02 */
        fprintf(stderr,"mcpgath:  no room in filesystem to copy %s for task %d.\n", 
                    fdp->name, my_task_number);  /* M02 */
        /* M03 fixed spacing in message */
        MPI_Abort(MPI_COMM_WORLD,-5); /* stop all other tasks */
    }
      k = fdp->content_size;  /*  M02 write amount = amount sent */
  
      if((j=fwrite(contentpr,sizeof(char),k,fp_out))!=k) {  /* M02 */
            fprintf(stderr," mcpgath: Failed writing of %s from task %d\n",
                    fdp->name,status.MPI_SOURCE); /*  M02 write failure */
            fprintf(stderr, "mcpgath: Wrote %d bytes while writing %d\n",j,k);
            MPI_Abort(MPI_COMM_WORLD,-8); /* stop all other tasks */
      }

  } while (tasks != number_of_tasks);  /* continue receiving messages until last msg sent */

  for (i = 0; i < FOPEN_MAX-1; i++)   
     if (file_tbl[i].ptr != NULL) fclose(file_tbl[i].ptr); 
                                          /* close the output files */

} /* End of receive side logic for tasks 1 thru N */
free(dest_dir);
free(rec_buffer);  /* M02 */
MPI_Finalize();
} /* end of main */
