/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 1995,2012          */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
/*
 *   File name: mcpscat.c
 *
 * History:
 *      03/27/98 AJI M01 changes to #define BLOCK_SIZE (D38675)
 *      05/22/98 AJI M02 changes to dynamic blocksize (D39519)
 *      06/17/98 PFC M03 fix labels on messages added by defects 38675 and
 *                       39519 (D41184)
 *      06/18/98 SVC M04 stdlib.h needed for getenv() (D40375)
 *      07/06/98 AJI M05 Routine needs multiple enhancements from counting # of flags
 *                       and multiple messages issued when file already exists (D34888)
 *      10/12/98 PFC M06 Corrected task number in msg issued for creating a file (D44487)
 *      02/03/99 PFC M07 check for DFS f/s when checking space (D46200)
 *      03/11/99 SVC M08 OS/390: move to mohonk, undo M07 (no DFS for OS/390) (D47705)
 *      03/15/99 SVC M09 OS/390: issue error message once when filesystem on home node gets full (D48900)
 *      10/15/03 PFC M10 allocate glob output structure prior to call (D100085)
 *      04/20/05 ECL M11 Linux port (F118848/D120380)
 *
 *
 *   Origins:
 *                   Copyright IBM Corporation 1995,2002
 *
 *                         All Rights Reserved
 *
 *   Permission to use, copy, modify, and distribute this software and its
 *   documentation for any purpose and without fee is hereby granted,
 *   provided that the above copyright notice appear in all copies and that
 *   both that copyright notice and this permission notice appear in
 *   supporting documentation, and that the name of IBM not be
 *   used in advertising or publicity pertaining to distribution of the
 *   software without specific, written prior permission.
 *
 *   Information: mcpscat does a parallel file copy scatter, taking a number    
 *   of files from task 0 and scatters them in sequence 
 *   in a round robin order of the number of tasks involved.
 * 
 *   The syntax is:
 *
 *      mcpscat [-f] [-i] source  .... destination  [POE options]   where
 *
 *         -f is an optional flag that says that the first file contains the
 *            names of the source files that should be scattered.  Each file
 *            name, in the file, must be specified on a separate line.  No
 *            wildcards are supported when this option is used.  Directory
 *            names are not supported in the file either.
 *            The mcpscat parameters should consist of a single source file name
 *            (for the list of files) and a destination directory when this
 *            option is used.  The files will then be scattered just as if they
 *            had all been specified on the command line in the same order as
 *            they are listed in the file.
 *
 *         -i is an optional flag when specified denotes to check for duplicate
 *            or existing files of the same name, and to not replace any existing
 *            file found.  If this flag is not specified, the default action is
 *            to replace any existing files with the source file.
 *
 *         source  can be one of the following:
 *             - a single file name, in which case that file is copied to all tasks. 
 *             - a single file name that contains a list of file names (-f option).
 *             - two or more file names, where those files will be distributed in a   
 *               round robin order to the tasks.
 *             - an expansion of file names, using wildcards, where the files will be 
 *               distributed in a round robin order to the tasks.
 *             - a directory name, where all files in that directory are copied in a 
 *               round robin order to the tasks.  
 *
 *         destination  is an existing destination directory name where the data will be
 *            copied to.
 *
 *         Wildcard characters must be enclosed in double quotes.
 *
 *         POE options  are any valid POE command line argument.  You can also
 *            specify any valid POE environment variable.
 *
 *   The number of nodes to move the file to (-procs), and the message
 *   passing protocol used to move the file (-euilib) are the POE
 *   options which may be of most interest to the user. See the IBM
 *   Parallel Environment for AIX Operation and Use guide for 
 *   information on POE options.
 *
 *   A default of 100K data block size is used for copying the data.  This can be 
 *   changed by setting the enviornment variable MP_BLKSIZE between 1 and 8000000.
 *
 *   return codes:  -1 : invalid number of arguments specified
 *                  -2 : invalid option flag specified      
 *                  -3 : unable to resolve input file name(s)       
 *                  -4 : Could not open input file for read                  
 *                  -5 : no room on destination node's filesystem                 
 *                  -6 : error opening file output file    
 *                  -7 : error creating output file       
 *                  -8 : MPI_Send of data failed              
 *                  -9 : Final MPI_Send failed              
 *                  -10 : MPI_Recv failed              
 *                  -11 : Failed opening temporary file
 *                  -12 : Failed writing temporary file
 *                  -13 : error renaming temp file to filename
 *                  -14 : input file is empty (zero byte file size)
 *                  -15 : invalid block size
 *                  -16 : error allocating storage
 *                  -17 : no of tasks and files are not compatible
 *                  -18 : not enough memory for list of file names

 */

#pragma comment(copyright,"@(#) 5765-F83 (C) Copyright IBM Corp. 1995,2002\0")
#pragma comment(user,"@(#) 1.29 src/ppe/poe/samples/mpi/mcpscat.c, ppe.poe.samples, ppe_rbarlx2, rbarlx21216b 12/04/23 13:26:33")

#include <stdio.h>
#include <stdlib.h>      /* M04 */
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#ifndef __linux   /* M11: sys/vnode.h not available on Linux */
 #include <sys/vnode.h>
#endif
#include <glob.h>
#include <mpi.h>
#include <errno.h>  /* M07 */

#define MAX_BLOCK_SIZE 8000000    /* M01 maximum block size user can request */

main(argc,argv)
int argc;
char *argv[];
{
  glob_t globbuf;              /* structure for obtaining file names */
  char in_file[FILENAME_MAX];  /* name of file to be moved */
  char *dest_dir;              /* name of output destination directory */
  FILE *fp;                    /* file pointer for input file */
  FILE *fp_out;                /* file pointer for file being written */
  struct stat statb;           /* structure needed for stat() function */
  struct statvfs stata;        /* structure needed for statvfs() function */
#ifndef __linux                  /* M11: VDIR & st_type not available on Linux */
  #define STATB_STTYPE_IS_VDIR statb.st_type == VDIR
#else
  #define STATB_STTYPE_IS_VDIR S_ISDIR(statb.st_mode)
#endif
  int i,j,k;                   /* general purpose integers */
  int opened_file;             /* indicates if out file is open for writing */
  int c;
  int fd;
  MPI_Status status;

  struct file_def {            /* file data sent to each node */
    int  name_size;            /* size of file name */
    long  file_size;            /* size of the file */
    int  content_size;         /* size of file data in this send */
    char name[FILENAME_MAX];   /* file name */
    mode_t file_mode;          /* file mode (permisssions etc...) */
  } file_definition;
 
  char temp_filename[FILENAME_MAX]; /* temporary filename */
  char task0_filename[FILENAME_MAX]; /* Task 0 output file name */
  char *in_name; /* input file name */

  char *env_block_ptr;         /* M01 pointer to env var to get MP_BLKSIZE value */
  char  *contentps;            /* M02 pointer to file data on send side */
  char  *contentpr;            /* M02 pointer to file data on receive side */
  char  *send_buffer;          /* M02 pointer to send buffer */
  char  *rec_buffer;           /* M02 pointer to receive buffer */
  struct file_def *fdp;        /* M02  */

  long int blocksize = 100000; /* M01 default blocksize is used if MP_BLKSIZE is not set */

  int number_of_tasks;         /* # nodes (including sender) executing */
  int my_task_number;          /* task number of this node */
  int cur_task_number = 0;     /* current task number for doing round robin distribution */
  int numfiles = 0;
  int prev_numfiles = 0;
  int exist = 0;      
  int read_names = 0;           /* if 1, then read file names from file */
  char *new_line;
  int type = 0;
  int lasttype = 41;
  int source = 0;
  int nbytes,msg_rc;
  int err_rc = 0;
  int send_ctr, task, r;
  int first_time_send_buffer = 1;    /* M02 */
  int first_time_rec_buffer = 1;    /* M02 */
  char save_filename[FILENAME_MAX];    /* M05 */
  int errnosave = 0;  /* M07 */
   

/***********************************************/
/* Set up message passing environment, using   */
/* the number of tasks (procs) to set up the  */
/* task group.                                */
/***********************************************/

MPI_Init(&argc,&argv); /* initialize MPI environment */
MPI_Comm_rank(MPI_COMM_WORLD, &my_task_number); /* get task number */
MPI_Comm_size(MPI_COMM_WORLD,&number_of_tasks); /* get number of tasks */

/*******************************************************/
/* Parse arguments and check for -i and/or -f flag     */
/*******************************************************/

while (--argc > 0 && (*++argv)[0] == '-')
  while (c = *++argv[0])  /* an option was found ... */
     switch (c) {
       case 'i':
           exist = 1;   /* check for duplicate existing files */
           break;
       case 'f':
           read_names = 1;   /* read file names from file */
           break;
       default:
           fprintf(stderr,"mcpscat:  Illegal option %c \n", c);
           fprintf(stderr,"usage: mcpscat [-f] [-i] source ... destination\n"); 
           MPI_Abort(MPI_COMM_WORLD,-2); /* stop all other tasks */
     }
                     
if(argc < 2) {    /* M05 should be 1-n arguments to this prog */
      fprintf(stderr,"usage: mcpscat [-f] [-i] source ... destination\n"); 
                                  /* if not, issue error message */
      MPI_Abort(MPI_COMM_WORLD,-1); /* stop all other tasks */
}

/* Handle destination directory, since that must be the last parameter */
dest_dir = (char *) malloc(strlen(argv[argc-1])+1);   /*stg for output dir name*/
strcpy(dest_dir,argv[argc-1]);                      /* get output dir name */

strcpy(in_file,argv[0]);                      /* get input file name */

/***********************************************/
/* Check to see if env varible is set thru cmd */
/* line. If the environment variable was set   */
/* get the value of it. Check to see if block  */
/* size is with in limits, if so allocate the  */
/* storage based on the MP_BLKSIZE.            */
/***********************************************/
   
  if (first_time_send_buffer == 1) {    /* M02 allocate the data area once for every file */
     if ((env_block_ptr = getenv("MP_BLKSIZE")) != NULL) {  /* M02 */
        blocksize = atoi(env_block_ptr); /* M02 */
     }  /* M02 */
     if ((blocksize <= 0) || (blocksize > MAX_BLOCK_SIZE)) {    /* M02 */
        fprintf(stderr, "mcpscat: blocksize has to be between 1 and %ld\n", MAX_BLOCK_SIZE);
                                                                /* M02 */
                     /* M03 added function label to message */
        MPI_Abort(MPI_COMM_WORLD,-15); /* M01 stop all other tasks */
     }  /* M02 */
  
    /* M02 Allocate data area based on block size */
  
     if ((send_buffer = (char *) malloc(sizeof(file_definition) + blocksize )) == NULL) {
                                                       /*  M02 */
        fprintf(stderr,"mcpscat:  error allocating storage for send buffer \n"); /* M02 */
                     /* M03 added function label to message */
        MPI_Abort(MPI_COMM_WORLD,-16); /* M01 stop all other tasks */
      }   /* M02 */
        first_time_send_buffer +=1;  /* M02 */
   }     /* M02 */
     contentps = send_buffer + sizeof(file_definition); /* M02 */

if (read_names) {
  globbuf.gl_pathc = 0;  /* init count */
  globbuf.gl_pathv = (char **) malloc(50 * sizeof(char *));
  if (globbuf.gl_pathv == NULL) {
    fprintf(stderr,"mcpscat:  Not enough memory for list of file names.\n");
    MPI_Abort(MPI_COMM_WORLD,-18); /* stop all other tasks */
  }
  globbuf.gl_offs = 50;  /* room for 50 entries to start */
} else {
  stat(in_file,&statb);  /* Check to see if source is a directory */
  if ( STATB_STTYPE_IS_VDIR ) strcat(in_file,"/*");
}

if(my_task_number == 0) {   /* Beginning of send side logic on task 0 */

  if (read_names) {
    if ( (fp = fopen(in_file, "r")) == NULL) {
      fprintf(stderr,"mcpscat:  Unable to open input file %s to read list of file names.\n", in_file);
      MPI_Abort(MPI_COMM_WORLD,-4);
    }
    while ( fgets(temp_filename, FILENAME_MAX, fp) != NULL) {
      if (globbuf.gl_pathc == globbuf.gl_offs) {
        /* time to get more space */
        if ( (globbuf.gl_pathv=realloc(globbuf.gl_pathv,
                            (globbuf.gl_offs + 50) * sizeof(char *))) == NULL) {
          /* can't get more space */
          fprintf(stderr,"mcpscat:  Not enough memory for list of file names.\n");
          MPI_Abort(MPI_COMM_WORLD,-18); /* stop all other tasks */
        }
        globbuf.gl_offs += 50;   /* increase max count  */
      }
      new_line = strchr(temp_filename, '\n');
      if (new_line != NULL)
        *new_line = '\0';   /* change new-line to Null char */
      globbuf.gl_pathv[globbuf.gl_pathc] = (char *) malloc(strlen(temp_filename)+1);
      if (globbuf.gl_pathv[globbuf.gl_pathc] == NULL) {
        /* can't get more space */
        fprintf(stderr,"mcpscat:  Not enough memory for list of file names.\n");
        MPI_Abort(MPI_COMM_WORLD,-18); /* stop all other tasks */
      }
      strcpy(globbuf.gl_pathv[globbuf.gl_pathc], temp_filename); /* copy */
      globbuf.gl_pathc++;   /* bump count */
    }
    fclose(fp);
  } else {
    /* Make first call to obtain file names */
    memset(&globbuf, 0, sizeof globbuf); /* M10 allocate glob structure area */
    glob(in_file,GLOB_DOOFFS,NULL,&globbuf);
    if (globbuf.gl_pathc <= 0) {   /* glob failed  */
      fprintf(stderr,"mcpscat:  Unable to resolve list of files for %s.\n",in_file);
      MPI_Abort(MPI_COMM_WORLD,-3); /* stop all other tasks */
    }
    for (i=0; i < argc-2; i++) {    /* If there are several source file names
                                     go get them.  */
      strcpy(in_file,argv[i+1]);                      
      stat(in_file,&statb);  /* Check to see if source is a directory */
      if ( STATB_STTYPE_IS_VDIR ) strcat(in_file,"/*");

      /* Input file name can either be an explicitly specified name
         of a file, a directory, or a "fuzzy" representation of files using
         meta characters (such as *.o) as long as the string is enclosed in
         double quotes.  We will use glob to get the list of real file names. */

      prev_numfiles = globbuf.gl_pathc;
      glob(in_file,GLOB_APPEND|GLOB_DOOFFS,NULL,&globbuf);
      if (globbuf.gl_pathc == prev_numfiles ) {   /* glob failed  */
        fprintf(stderr,"mcpscat:  Unable to resolve list of files for %s.\n",in_file);
        MPI_Abort(MPI_COMM_WORLD,-3); /* stop all other tasks */
      }
    }
  }
  numfiles = globbuf.gl_pathc;
  i = 0;

  while(i < numfiles) { /***  BEGINNING OF ROUND ROBIN LOOP FOR DISTRIBUTING FILES  ***/

   if (numfiles > 1) {  /* M02 */
        if(number_of_tasks > numfiles) {  /* M02 */
          fprintf(stderr,"mcpscat:  the number of files %d is less than number of tasks %d \n", numfiles,
                  number_of_tasks); /* M02 tasks and files are not compatible */
                     /* M03 added function label to message */
          MPI_Abort(MPI_COMM_WORLD,-17); /* M02 stop all other tasks */
         }
     }

   /* For each file we have, begin sending files one at a time in a 1 to 1 
      correspondence to the tasks involved.  If the number of files exceeds the
      number of tasks, go thru the tasks again sending the remaining files
      continuing until all files have been sent.    */

    file_definition.name_size = strlen(globbuf.gl_pathv[i]); /* get len file */
    strcpy(file_definition.name,globbuf.gl_pathv[i]); /* get name of file */

    /***********************************************/
    /* if no information is available about the    */
    /* file or the file cannot be opened...        */
    /***********************************************/

    if(stat(globbuf.gl_pathv[i],&statb) == -1 || ! (fp=fopen(globbuf.gl_pathv[i],"r"))) {
      fprintf(stderr,"mcpscat: Could not open file %s for reading\n",globbuf.gl_pathv[i]);
      MPI_Abort(MPI_COMM_WORLD,-4); /* stop all other tasks */
    }

    /* Check to see if file is a directory.  If so, it must be skipped. */

    if ( STATB_STTYPE_IS_VDIR ) { 
      i++;   /* process next file  */
      fclose(fp);  /* close this file  */
      continue; /* continue with next file */
    }

    if (statb.st_size == 0) {     /* File is empty....issue msg */
      fprintf(stderr,"mcpscat: File %s is empty\n",globbuf.gl_pathv[i]);
      fclose(fp);  /* close this file  */
      MPI_Abort(MPI_COMM_WORLD,-14); /* stop all other tasks */
    }

    file_definition.file_mode = statb.st_mode;  /* get file mode */
    file_definition.file_size = statb.st_size;  /* get file len from struct */

    /***********************************************/
    /* If this is the task with the originating    */
    /* file, we can write it directly to its       */
    /* destination directory and not bother with   */
    /* message passing.                            */
    /***********************************************/

    opened_file = 0;                       /* file not yet opened for this task */
    if(cur_task_number == 0)    /* if this is task 0 ..... open file to write */
      if (opened_file == 0) {  /* Output file not open yet.... */

     /************************************************/
     /* Verify that there is space for the file.     */
     /* Divide the file size by the block size of    */  
     /* the file system. If that value is > the      */
     /* number of free blocks in the file system,    */
     /* stop processing this task and return.        */
     /************************************************/
 
          strcpy(task0_filename,dest_dir);  /* Replace name with directory and file name */ 
          strcat(task0_filename,"/"); 
          in_name = strrchr(file_definition.name,'/');  /* Strip off path name */
          in_name ? strcat(task0_filename,++in_name) : 
                    strcat(task0_filename,file_definition.name);  
                                          /* Join directory name and file name, after 
                                              stripping off input path name */
          if(!(statvfs(task0_filename, &stata))) {

      /* M07 Check file system space only if file sys type is not DFS */

#ifndef __linux  /* M08/M11: No DFS support on Linux */
            if (strncmp(stata.f_basetype,"dfs",3) != 0)    /* M07 */
#endif       /* M08 */
              if (stata.f_bavail<((unsigned long)file_definition.file_size/stata.f_frsize+1)) {
                      fprintf(stderr,"mcpscat:no room in filesystem to copy %s for task %d.\n", 
                      task0_filename, my_task_number);
                MPI_Abort(MPI_COMM_WORLD,-5); /* stop all other tasks */
              }
           /* file exists ... if -i flag specified, we do not want to replace it.     */

            if (exist == 1)      /* Do not replace existing file, issue msg and continue */
              fprintf(stderr,"mcpscat: File %s already exists on task 0, and will not be copied.\n",
                       task0_filename);
            else  {   /* We can replace an existing file (-i not specified) */
               if( !(fp_out = fopen(task0_filename,"w")) ) { /* open "to" file */
                 fprintf(stderr," mcpscat: Failed on open of %s on task %d\n",
                         task0_filename,cur_task_number);
                 MPI_Abort(MPI_COMM_WORLD,-6); /* stop all other tasks */
               }
               opened_file = 1;            /* "to" file on originating opened */
            }
          }
          else {                  /* statvfs failed, file must not exist,
                                    so create it. */
             fd = creat(task0_filename,file_definition.file_mode);
             fp_out = fdopen(fd,"w");
             if (fp_out == NULL) {   /* open failed.... */
              fprintf(stderr,"mcpscat: Failed create of output file %s for task %d \n",
                       task0_filename,cur_task_number); /* M06 - correct task number in msg */
              MPI_Abort(MPI_COMM_WORLD,-7); /* stop all other tasks */
             }
             opened_file = 1;                      /* "to" file on originating opened */
          }
    }        /* end if task = 0, opening file for write */

              
   do {     /* read contents of the file */
     
   if(statb.st_size > blocksize)  {   /* M01 if # bytes left in file > block */
          k = blocksize;                  /* M01 write amount = 1 block */
        }
        else k = statb.st_size; /* if # bytes left < or = block */
                                /* write amount = amount left */
        statb.st_size -= k;                /* decrement amount left to write */

        if( !fread(contentps,sizeof(char),k,fp)) { /* M02 read piece */
          fprintf(stderr,"mcpscat Attention: fread of %s may have failed\n",
                  file_definition.name);
        }
        file_definition.content_size = k;  /* # bytes read in structure sent */
   
        if(cur_task_number == 0 ) {  
                 /* if this is task 0 ..... copy file directly */

          if (opened_file == 1)
            if((j=fwrite(contentps,sizeof(char),k,fp_out))!=k) {  /* M02 */
               fprintf(stderr," mcpscat: Failed on write of %s on task %d\n",
                   file_definition.name,cur_task_number); /* write failure */
               fprintf(stderr, "mcpscat: Wrote %d bytes while writing %d\n",j,k);
#ifndef __linux /* M09/M11 */
               err_rc = 2;
               continue;     /* continue, but save error code   */
#else       /* M09/M11 */
               MPI_Abort(MPI_COMM_WORLD,-5); /* M09: stop all other tasks */
#endif      /* M09/M11 */

          }
          if (numfiles == 1)    /* This takes care of the case where there is only one
                                   file, so we want to copy it to all other tasks. */ 
            for (r = 1; r < number_of_tasks; r++) {
               memcpy(send_buffer, (char *) &file_definition, sizeof (file_definition)); /* M02 */
               msg_rc = MPI_Send(send_buffer, sizeof(file_definition)+blocksize,MPI_CHAR,
                        r,type,MPI_COMM_WORLD); /* M02 */
              if (msg_rc) {
                  fprintf(stderr," mcpscat: MPI_Send of %s to task %d failed.\n",
                      file_definition.name,r); /* write failure */
                  MPI_Abort(MPI_COMM_WORLD,-8); /* stop all other tasks */
              }
           }   /* end loop of sends */
        }
        else {   /* reading a file for task 1 thru task N, send it to that task... */

          send_ctr = 1;
          if (numfiles == 1) send_ctr = number_of_tasks;
          for (r = 1; r <= send_ctr; r++) {

              if (numfiles == 1) task = r;
              else task = cur_task_number;
              memcpy(send_buffer, (char *) &file_definition, sizeof (file_definition)); /* M02 */
              msg_rc = MPI_Send(send_buffer, sizeof(file_definition)+blocksize,MPI_CHAR,
                       task,type,MPI_COMM_WORLD);
              if (msg_rc) {
                  fprintf(stderr," mcpscat: MPI_Send of %s to task %d failed.\n",
                      file_definition.name,task); /* write failure */
                  MPI_Abort(MPI_COMM_WORLD,-8); /* stop all other tasks */
              }
           }   /* end loop of sends */
        }

    } while (statb.st_size > 0); /* doing read of the input file */

    if (cur_task_number == number_of_tasks - 1) cur_task_number = 0;   
                             /* Yes, a file has been sent to each task...    */
                        /* Reset current task number to begin another round */

    else cur_task_number++;  /* No, still more tasks to send to, go to next task */
    i++;
    fclose(fp);                      /* close the file we've been reading */
    if (opened_file) {
      /* clean up task 0 writing   */
      fclose(fp_out);                /* close the file we've been writing */
      chmod(file_definition.name,file_definition.file_mode); /* set file mode */
    }
  }   /* end of round robin loop */
    for (i=1; i<number_of_tasks; i++) {
      msg_rc = MPI_Send(&i,sizeof(i),MPI_CHAR,i,lasttype,MPI_COMM_WORLD);
      if (msg_rc) { /* write failure */
         fprintf(stderr," mcpscat: last MPI_Send to task %d failed.\n",i); 
         MPI_Abort(MPI_COMM_WORLD,-9); /* stop all other tasks */
      }
    }

 }                                             /* end of "if this is task 0" */
/************** END OF SEND LOGIC ***********************************************/
    
else {      /* Receive side of logic for Tasks 1 thru N   */

/*****************************************************/
/* This is the receive side logic, for receiving     */
/* the files, on tasks 1 thru N.  Task 0 is handled  */
/* on the send side, by writing the file directly to */
/* the destination.                                  */
/* first we will allocate receive buffer, and then   */
/* we will do a receive, open the file for writing,  */
/* and write the block, until the entire file has    */
/* been received.                                    */
/*****************************************************/
    
  opened_file = 0;  /* set file not opened flag  */
  do {   /* Receive messages until we get a message of a particular type, that   
            signifies the last file has been sent */
  do {                                   /*cont reading and writing until eof*/
   if (first_time_rec_buffer == 1) { /* M02 */
      if ((rec_buffer = (char *) malloc(sizeof(file_definition)+blocksize)) == NULL) { /* M02 */
         fprintf(stderr,"mcpscat:  error allocating recieve buffer storage\n");  /* M02 */
         MPI_Abort(MPI_COMM_WORLD,-16); /* M02 stop all other tasks */
                     /* M03 added function label to message */
       }   /* M02 */
      first_time_rec_buffer +=1; /* M02 */
    }  /* M02 */
   fdp = (struct file_def *)rec_buffer;  /* M02 */
   msg_rc = MPI_Recv(rec_buffer,sizeof(file_definition)+blocksize,MPI_CHAR,source,
            MPI_ANY_TAG,MPI_COMM_WORLD,&status);  /* M02 */
   if (msg_rc) {
       fprintf(stderr," mcpscat: MPI_Recv of %s on task %d failed.\n",
                   fdp->name,my_task_number); /* M02 write failure */
       MPI_Abort(MPI_COMM_WORLD,-10); /* stop all other tasks */
    }
   contentpr = rec_buffer + sizeof(file_definition); /* M02 */

   if (status.MPI_TAG == lasttype) break;    
   if(opened_file == 0) {             /*if output file has not been opened*/

      /************************************************/
      /* Verify that there is space for the file.     */
      /* Divide the file size by the block size of    */  
      /* the file system. If that value is > the      */
      /* number of free blocks in the file system,    */
      /* stop processing this task and return.        */
      /************************************************/
 
        strcpy(task0_filename,dest_dir);  /* Replace name with directory and file name */ 
        strcat(task0_filename,"/"); 
        in_name = strrchr(fdp->name,'/');  /* M02 Strip off path name */
        in_name ? strcat(task0_filename,++in_name) : 
                  strcat(task0_filename,fdp->name);  
                                          /* M02 Join directory name and file name, after 
                                              stripping off input path name */
        strcpy(fdp->name,task0_filename);  /* M02 Replace name with directory and file name */ 
        if(!(statvfs(fdp->name, &stata))) {  /* M02 */
         /* file exists ... if -i flag specified, we do not want to replace it.     */

          if (exist == 1) {    /* Do not replace existing file, issue msg and continue */
             if (strcmp( save_filename,fdp->name) != 0) {  /* M05 compare both filename */ 
                strcpy(save_filename,fdp->name);           /* M05 */  
                fprintf(stderr,"mcpscat: File %s already exists on task %d, and will not be copied.\n",
                       fdp->name,my_task_number); /* M05 */
             } /* M05 */

            /* Without an open file, it is difficult to keep track of how   */
            /* many bytes to write to this file.  So, set the size value    */
            /* controlling the loop to zero and let the bigger status.MPI   */
            /* loop handle things.                                          */
            statb.st_size = 0;
            continue;          /* continue with next file   */
          }

      /* M07 Check file system space only if file sys type is not DFS */

#ifndef __linux  /* M08/M11: No DFS support on Linux */
          if (strncmp(stata.f_basetype,"dfs",3) != 0)    /* M07 */
#endif       /* M08/M11 */
            if (stata.f_bavail<((unsigned long)fdp->file_size/stata.f_frsize+1)) {  /* M02 */
              fprintf(stderr,"mcpscat:  no room in filesystem to copy %s for task %d.\n", 
                      fdp->name, my_task_number);  /* M02 */
                     /* M03 fixed spacing in message */
              MPI_Abort(MPI_COMM_WORLD,-5); /* stop all other tasks */
            }
        }
        else {                  /* statvfs failed, file must not exist,
                                  so create it. */
           fd = creat(fdp->name,fdp->file_mode);  /* M02 */
        }
       /* Create a temporary file to write data to, in case we encounter
          a problem down the line.  This way, if there is an error, the
          original data is preserved, for files that already exist and
          we are overwriting.     */
 
        sprintf(temp_filename,"%s.%d\0",fdp->name, getpid());  /* M02 */ 
        if( !(fp_out = fopen(temp_filename,"w")) ) {    /* open it */
          fprintf(stderr," mcpscat: Failed opening temp file %s on task %d\n",
                  temp_filename,my_task_number);
          MPI_Abort(MPI_COMM_WORLD,-11); /* stop all other tasks */
        }
        opened_file = 1;                 /* indicate file has been opened */
        statb.st_size = fdp->file_size; /* M02 get size of file  */
    }    /* end if output file has not been opened.... */

  /* Check filesystem size to ensure there is enough room to hold data */

    statvfs(fdp->name, &stata);  /* M02 */

    /* M07 Check file system space only if file sys type is not DFS */

#ifndef __linux  /* M08/M11: No DFS support on Linux */
    if (strncmp(stata.f_basetype,"dfs",3) != 0)    /* M07 */
#endif       /* M08/M11 */
      if (stata.f_bavail<((unsigned long)fdp->file_size/stata.f_frsize+1)) {  /* M02 */
          fprintf(stderr,"mcpscat:  no room in filesystem to copy %s for task %d.\n", 
                      fdp->name, my_task_number);  /* M02 */
                     /* M03 fixed spacing in message */
          MPI_Abort(MPI_COMM_WORLD,-5); /* stop all other tasks */
      }

    k = fdp->content_size;  /* M02 write amount = amount sent */

    if((j=fwrite(contentpr,sizeof(char),k,fp_out))!=k) {  /* M02 */
        fprintf(stderr," mcpscat: Failed writing temp file %s on task %d\n",
                temp_filename,my_task_number); /* write failure */
        fprintf(stderr, "mcpscat: Wrote %d bytes while writing %d\n",j,k);
        MPI_Abort(MPI_COMM_WORLD,-12); /* stop all other tasks */
    }
    statb.st_size -= k;              /* decrement # bytes left to write  */

  } while (statb.st_size > 0);       /* cont above until no more data in file */

  if (opened_file ==1) {
    fclose(fp_out);                  /* close the file we've been writing */
    chmod(temp_filename,fdp->file_mode); /*  M02 set file mode */
    if(rename(temp_filename,task0_filename) != 0) { /* rename temp file */
        fprintf(stderr, "mcpscat: Error renaming temp file %s to %s\n",
                temp_filename, task0_filename);
        MPI_Abort(MPI_COMM_WORLD,-13); /* stop all other tasks */
    }
    opened_file = 0;
  }  /* end rename temp file */
  } while (status.MPI_TAG != lasttype);    /* continue receiving messages until last msg sent */
 } /* End of receive side logic for tasks 1 thru N */
if ( my_task_number != 0 ) free(rec_buffer);  /* M02/M11 */
MPI_Barrier(MPI_COMM_WORLD);
MPI_Finalize();
exit(err_rc);
} /* end of main */
